#include "hw.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define _BV(bit) (1 << (bit)) // _BV macro is not defined for atmega8535 so define it here

// BTNx is 1 if respective button is pressed
#define BTN0() (bit_is_clear(PINB, 0))
#define BTN1() (bit_is_clear(PINB, 1))
#define BTN2() (bit_is_clear(PINB, 2))

#define BUTTON_PRESS_DELAY_MS 220

// display segments configuration
#define _s_A 0
#define _s_B 1
#define _s_C 2
#define _s_D 3
#define _s_E 4
#define _s_F 5
#define _s_G 6
#define _s_dot 7

const unsigned char segs[] =
{
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F), //0
	_BV(_s_B) | _BV(_s_C), //1
	_BV(_s_A) | _BV(_s_B) | _BV(_s_D) | _BV(_s_E) | _BV(_s_G), //2
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_G), //3
	_BV(_s_B) | _BV(_s_C) | _BV(_s_F) | _BV(_s_G), //4
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G), //5
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //6
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C), //7
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G),//8
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_F) | _BV(_s_G),//9

	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //A
	_BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //B
	_BV(_s_A) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F), //C
	_BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_G), //D
	_BV(_s_A) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //E
	_BV(_s_A) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //F
	_BV(_s_A) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E) | _BV(_s_F), //G
	_BV(_s_B) | _BV(_s_C) | _BV(_s_E) | _BV(_s_F) | _BV(_s_G), //H 17
	_BV(_s_B) | _BV(_s_C) | _BV(_s_D) | _BV(_s_E), //J 18
	_BV(_s_A) | _BV(_s_D), //: 19
	_BV(_s_A) | _BV(_s_B) | _BV(_s_C) | _BV(_s_D) //) 20
};

volatile unsigned char hour = 0;
volatile unsigned char min = 0;
volatile unsigned char sec = 0;
volatile unsigned char sec1 = 0;
volatile unsigned char sec10 = 0;
volatile unsigned char min1 = 0;
volatile unsigned char min10 = 0;
volatile unsigned char hour1 = 0;
volatile unsigned char hour10 = 0;
volatile unsigned char activeseg = 0; // indicates active display segment (used for multiplexing)
volatile unsigned char activeled = 0; // indicates active binary led (used for multiplexing)
volatile unsigned int timer0ignore = 0;
volatile unsigned int timestarted = 0;

static void refreshBackendTimeVariables()
{
	sec1 = sec % 10;
	sec10 = sec / 10;
	min1 = min % 10;
	min10 = min / 10;
	hour1 = hour % 10;
	hour10 = hour / 10;
}

int main(void)
{
	// set all PORTA, PORTC and PORTD pins as output
	DDRA = 0xFF;
	DDRC = 0xFF;
	DDRD = 0xFF;
	DDRB &= ~_BV(PB0); // input: button
	DDRB &= ~_BV(PB1); // input: button
	DDRB &= ~_BV(PB2); // input: button
	DDRB |= _BV(PB4); // output: LED anode "1"
	DDRB |= _BV(PB5); // output: LED anode "2"
	DDRB |= _BV(PB6); // output: LED anode "4"
	DDRB |= _BV(PB7); // output: LED anode "8"

	// show empty display/leds at the beginning
	PORTA = 0x00;
	PORTC = 0xFF;
	PORTD = 0xFF;
	PORTB &= ~0b11110000;

	// PORTA: diodes (common anodes)
	// PORTB: buttons and diodes (cathodes)
	// PORTC: display (common anodes)
	// PORTD: segments (cathodes)
	
	// TIMER0 (led multiplexing) configuration
	TIMSK |= _BV(TOIE0); // enable overflow interrupt
	TCCR0 |= _BV(CS01); // configure prescaler to 8 and start timer

	// TIMER2 configuration (RTC)
	TIMSK |= _BV(TOIE2); // enable overflow interrupt
	TCCR2 |= _BV(CS22) | _BV(CS20); // configure prescaler to 128 and start timer (because 32768/256 = 128) - 256 is max. value of 8-bit timer2
    ASSR |= _BV(AS2); // enable asynchronous mode

	// TIMER1 configuration
	TCCR1B |= _BV(WGM12); // configure timer 1 for CTC mode
	TIMSK |= _BV(OCIE1A); // enable CTC interrupt
	OCR1A = 4518; // set CTC compare value, this is to make the display not blinking to human eye (less than 10000)
	TCCR1B |= _BV(CS10); // configure prescaler to 1 and start timer
	
  	sei(); // enable global interrupts
	  
	// show AHOJ :) then wait a little, and continue
	hour10 = 10; // A
	hour1 = 17; // H
	min10 = 0; // O
	min1 = 18; // J
	sec10 = 19; // :
	sec1 = 20; // )
	_delay_ms(5000);

	hour10 = hour1 = min10 = min1 = sec10 = sec1 = 0;
	timestarted = 1;
	
    while (1)
    {
		if (BTN0())
		{
			hour++;
			
			if (hour == 24)
			{
				hour = 0;
			}
			
			refreshBackendTimeVariables();
			_delay_ms(BUTTON_PRESS_DELAY_MS);
		}
		if (BTN1())
		{
			min++;
			
			if (min == 60)
			{
				min = 0;
			}
			
			refreshBackendTimeVariables();
			_delay_ms(BUTTON_PRESS_DELAY_MS);
		}
		if (BTN2())
		{
			sec++;
			
			if (sec == 60)
			{
				sec = 0;
			}
			
			refreshBackendTimeVariables();
			_delay_ms(BUTTON_PRESS_DELAY_MS);
		}
    }
}

// ISR of interrupt that is called once per second
// increases internal time information by 1 second each time
ISR(TIMER2_OVF_vect)
{
	if (timestarted == 1 && !BTN2())
	{
		sec++;

		if (sec == 60)
		{
			sec = 0;
			min++;
		}
	
		if (min == 60)
		{
			min = 0;
			hour++;
		}
	
		if (hour == 24)
		{
			hour = 0;
		}
		
		refreshBackendTimeVariables();
	}
}

// display multiplexing
ISR(TIMER1_COMPA_vect)
{
	if (activeseg == 5)
	{
		activeseg = 0;
	} else {
		activeseg++;
	}

	switch (activeseg)
	{
		case 0:
			PORTD = ~segs[hour10];
			PORTC = 1;
		break;
		case 1:
			PORTD = ~segs[hour1];
			PORTC = 2;

			// blinking double dot
			if (timestarted == 1)
			{
				if ((sec1 & 1) == 0)
				{
					PORTD &= ~_BV(_s_dot);
				} else {
					PORTD |= _BV(_s_dot);
				}
			}
		break;
		case 2:
			PORTD = ~segs[min10];
			PORTC = 4;
		break;
		case 3:
			PORTD = ~segs[min1];
			PORTC = 8;
		break;
		case 4:
			PORTD = ~segs[sec10];
			PORTC = 16;
		break;
		case 5:
			PORTD = ~segs[sec1];
			PORTC = 32;
		break;
	}
}

// binary leds multiplexing
ISR(TIMER0_OVF_vect)
{
	timer0ignore++;

	// emulate custom prescaler to avoid slightly noticeable light-up of diodes that should be off
	if (timer0ignore % 5 == 0)
	{
		if (timestarted == 1)
		{
			if (activeled == 5)
			{
				activeled = 0;
			} else {
				activeled++;
			}

			switch (activeled)
			{
				case 0:
					PORTA = _BV(PA0);
					PORTB = ~(hour10 << 4);
				break;
				case 1:
					PORTA = _BV(PA1);
					PORTB = ~(hour1 << 4);
				break;
				case 2:
					PORTA = _BV(PA2);
					PORTB = ~(min10 << 4);
				break;
				case 3:
					PORTA = _BV(PA3);
					PORTB = ~(min1 << 4);
				break;
				case 4:
					PORTA = _BV(PA4);
					PORTB = ~(sec10 << 4);
				break;
				case 5:
					PORTA = _BV(PA5);
					PORTB = ~(sec1 << 4);
				break;
			}
		}
	}
}
